//------------------------------------
//  (c) Reliable Software, 1996, 97
//------------------------------------

//
// Plane
//
void PlaneSet(PLANE *P, RPOINT *center, VECTOR *perpendicular, OPTICALPROP *pr, COLORREF *pixels, int width, int height)
{
  VECTOR Temp = {0,0,1};

  P->Obj.Type = OBJECT_PLANE;
  P->Obj.OpticalProperty = *pr;

  P->Center = *center;
  P->Pixels = pixels;
  P->Width  = width;
  P->Height = height;

  P->Normal = *perpendicular;
  NormalizeVector(&P->Normal);

  P->Cos_inv = 1.0 / GetDotProductOfVectors(&P->Normal, &Temp);
}

// Vector normal at intersection
VECTOR PlaneNormal(PLANE *P, RPOINT *pt)
{
  return P->Normal;
}

void PlaneGetPixel(PLANE *P, RPOINT *pt, COLOR *color)
{
  int x,y;
  COLORREF rgb;

	// texture mapping
	// convert to local plane coordinates
	x = (int)ceil (pt->x * P->Cos_inv);
	y = (int)ceil (pt->y);
	// divide them modulo tile size

	x %= P->Width;
//	x = (x-1) & (P->Width -1);
	if (x < 0)
	    x += P->Width;

	y %= P->Height;
//	y = (y-1) & (P->Height - 1);
	if (y < 0)
	    y += P->Height;
	// find the corresponding pixel in the tile
	rgb = P->Pixels [y * P->Width + x];
	SetColorFromRGB(color, rgb);
}

bool PlaneIntersect(PLANE *P, RAY* pRay)
{
  double inc,t;
  VECTOR oc;
  // ( org + t * dir - center ) * normal = 0

  // cosine of incidence angle
  inc = GetDotProductOfVectors(&pRay->Direction, &P->Normal);
  if ( inc >= EPSILON )
	  return false;

  SetVectorFromPoints(&oc, &P->Center, &pRay->Origin);

  t = GetDotProductOfVectors(&oc, &P->Normal);
  t /= inc;
  if ( t > EPSILON ){
    RayRecord(pRay, t, (VISIBLEOBJECT*)P);
    return true;
  }
  return false;
}

Shader *PlaneCreateShader(PLANE *P, RAY *pRay)
{
  Shader *pShader;

  //pShader = (Shader*)malloc(sizeof(Shader));
  pShader = (Shader*)HeapAlloc(g_hHeap, 0, sizeof(Shader));
  ShaderSet(pShader, SHADER_PLANE, pRay, (VISIBLEOBJECT*)P);
  return pShader;
}


