// Reliable Software converted this listing from
// a Java listing into C++.
//
// Russell Hayward converted the C++ code to C
//
// I made a few minor alterations to
// Russells code. (John Findlay)
//
//------------------------------------
//  (c) Reliable Software, 1996, 97
//------------------------------------

/*==============================================================

	NOTE: Only include this file 'RAYTRACER.C' in the project.

		  Compile with 'Optimize' on.

===============================================================*/

#include <windows.h>
#include <math.h>
#include <gc.h>

#undef MAKEWORD
#define MAKEWORD(l,h) WORD(BYTE(l) | (BYTE(h) << 8))

typedef int bool;
#define true 1
#define false 0
//---------------------------------------------------------------------------
#include "Wrappers.h"
#include "Objects.h"
#include "Ray.h"
#include "Shader.h"
#include <stdio.h>
#include "Plane.h"
#include "Sphere.h"
//---------------------------------------------------------------------------

long long ll_zero = 0LL;

void CreateScene(),DestroyScene();
void InitRayTracer();
DWORD WINAPI DrawingThreadEntryPoint(void*);
void DestroyRayTracer();

LRESULT CALLBACK WndProc(HWND,UINT,WPARAM,LPARAM);
//---------------------------------------------------------------------------
#define MaxX 500
#define MaxY 400
#define IDI_ICON 101

HWND 	MainWin;
HANDLE  g_hHeap = NULL;

int ThreadIsDying=0;
HANDLE hThread;
CRITICAL_SECTION CriticalSection;

HBITMAP hSceneBmp = NULL;

#define MAX_OBJ 6

RPOINT ViewPoint;	// Position of the eye
RPOINT LightPoint;	// Position of light source
COLOR Ambient;		// Added to diffused reflection
int   MaxDepth;		// Recursion depth

VISIBLEOBJECT *Objects[MAX_OBJ]; // List of objects
int   ObjCount;

COLORREF *PlanePixels = NULL;
//---------------------------------------------------------------------------
#include "Objects.c"
#include "Plane.c"
#include "Sphere.c"

#include "Ray.c"
#include "Shader.c"
//========================================

//---------------------------------------------------------------------------
int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, char *cmdParam, int cmdShow)
{
	char *className = "RayToy";
	char *caption = "Reliable Software's 'LCC' ToyRay Tracer (www.relisoft.com)";
	MSG  msg;
	WNDCLASS cl;

	cl.style			= CS_HREDRAW | CS_VREDRAW;
	cl.lpfnWndProc		= WndProc;
	cl.cbClsExtra		= 0;
	cl.cbWndExtra		= 0;
	cl.hInstance		= hInst;
	cl.hIcon			= LoadIcon(hInst, MAKEINTRESOURCE(1));
	cl.hCursor			= LoadCursor(0, IDC_ARROW);
	cl.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
	cl.lpszMenuName		= 0;
	cl.lpszClassName	= className;

	RegisterClass(&cl);

	MainWin=CreateWindow(className, caption, WS_SYSMENU | WS_MINIMIZEBOX,
						20, 20, MaxX, MaxY, NULL, NULL, hInst, NULL);

	// Using Heap allocation is faster than malloc()
	g_hHeap = GetProcessHeap();

	if (MainWin != NULL){
		InitRayTracer();
		CreateScene();
		ResumeThread(hThread);

		ShowWindow(MainWin, cmdShow);
		UpdateWindow(MainWin);

		while (GetMessage(&msg, NULL, 0, 0) >0 ){
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		DestroyRayTracer();
		DestroyScene();

		return msg.wParam;
	}

	return EXIT_FAILURE;

}
//---------------------------------------------------------------------------
// Scene
//---------------------------------------------------------------------------
void CreateScene()
{
	double Radius;
	RPOINT Center;
	OPTICALPROP Optics;
	SPHERE *S;
	int yy;

	MaxDepth=4;
	SetPoint(&ViewPoint, 300, 150, 600);
	SetPoint(&LightPoint, -500, -100, 200);
	SetColor(&Ambient, 0.2, 0.2, 0.2);
	ObjCount = 0;

	ZeroMemory(&Optics, sizeof(Optics));

	{// Plane
	HBITMAP bmp;

	bmp = (HBITMAP)LoadImage(NULL, "zlogo.bmp", IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
	if (bmp == NULL){
		MessageBox(MainWin,"Can't find zlogo.bmp", NULL, MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
	}else{
		BITMAP bm;
		HDC TempDC;
		int x,y;
		VECTOR Perpendicular;
		PLANE *P;

		GetObject(bmp, sizeof(BITMAP), &bm);
		PlanePixels = (COLORREF*)malloc(sizeof(COLORREF) * bm.bmWidth * bm.bmHeight);
		TempDC = CreateCompatibleDC(NULL);
		SelectObject(TempDC, bmp);
		for (y = 0; y < bm.bmHeight; y++){
			yy = y*bm.bmWidth;
			for (x = 0; x < bm.bmWidth; x++){
				PlanePixels[x + yy] = GetPixel(TempDC, x, y);
			}
		}
		DeleteDC(TempDC);
		DeleteObject(bmp);

		SetPoint(&Center, 500, 0, -400);
		SetVector(&Perpendicular, -1, 0, 2);
		SetColor(&Optics.Col, 1, 1, 1);
		Optics.Refractiveness = 0;
		Optics.Transparency = 0;
		Optics.Reflectiveness = 0;
		Optics.Diffusion  = 1;
		Optics.IsLuminous   = false;

		P = (PLANE*)malloc(sizeof(PLANE));
		PlaneSet(P, &Center, &Perpendicular, &Optics, PlanePixels, bm.bmWidth, bm.bmHeight);
		Objects[ObjCount++] = (VISIBLEOBJECT*)P;
	}
	}

	// Transparent Sphere
	SetPoint(&Center, 359, 149, -300);
	Radius = 200;
	SetColor(&Optics.Col, 1, 1, 1);
	Optics.Refractiveness = 1.5;
	Optics.Transparency = 0.6;
	Optics.Reflectiveness = 0.3;
	Optics.Diffusion  = 0;
	Optics.IsLuminous   = false;

	S = (SPHERE*)malloc(sizeof(SPHERE));
	SphereSet(S, &Center, Radius, &Optics);
	Objects[ObjCount++] = (VISIBLEOBJECT*)S;

	// Reflecting Sphere
	SetPoint(&Center, 79, 295, -108);
	Radius = 84;
	SetColor(&Optics.Col, 0.4, 0.95, 0.3);
	Optics.Refractiveness = 0;
	Optics.Transparency = 0;
	Optics.Reflectiveness = 0.5;
	Optics.Diffusion  = 0.4;
	Optics.IsLuminous   = false;

	S = (SPHERE*)malloc(sizeof(SPHERE));
	SphereSet(S, &Center, Radius, &Optics);
	Objects[ObjCount++] = (VISIBLEOBJECT*)S;

	// Light source
	Center = LightPoint;
	Radius = 120;
	SetColor(&Optics.Col, 1.1, 1.1, 0.9);
	Optics.Refractiveness = 0;
	Optics.Transparency 	= 0;
	Optics.Reflectiveness = 0.5;
	Optics.Diffusion 		= 0.5;
	Optics.IsLuminous 	= true;

	S = (SPHERE*)malloc(sizeof(SPHERE));
	SphereSet(S, &Center, Radius, &Optics);
	Objects[ObjCount++] = (VISIBLEOBJECT*)S;
}
//---------------------------------------------------------------------------
void DestroyScene()
{
	int i;

	for (i = 0; i < ObjCount; i++){
		HeapFree(g_hHeap, 0, Objects[i]);
		//free(Objects[i]);
	}
	HeapFree(g_hHeap, 0, PlanePixels);
	//free(PlanePixels);
}
//---------------------------------------------------------------------------
// RayTracer
//---------------------------------------------------------------------------
void InitRayTracer()
{
	DWORD ThreadId;
	HDC DC;

	InitializeCriticalSection(&CriticalSection);

	hThread = CreateThread(NULL, 0, DrawingThreadEntryPoint, NULL, CREATE_SUSPENDED, &ThreadId);

	DC = GetDC(MainWin);
	hSceneBmp = CreateCompatibleBitmap(DC, MaxX, MaxY);

	{ // Clear the bitmap
	HDC TempDC;
	RECT rc = {0, 0, MaxX, MaxY};

	TempDC = CreateCompatibleDC(DC);
	SelectObject(TempDC, hSceneBmp);

	FillRect(TempDC, &rc, (HBRUSH)GetStockObject(WHITE_BRUSH));

	DeleteDC(TempDC);
	}

	ReleaseDC(MainWin, DC);

}
//---------------------------------------------------------------------------

DWORD WINAPI DrawingThreadEntryPoint(void *pArg)
{
	HDC DC,TempDC;
	COLORREF Line[MaxX];
	RAY Root;
	RPOINT pixel;
	COLOR shade;
	VECTOR direction;
	int x, y = 0;
	int t;
	bool j;
	t = GetTickCount();
	j = false;

	DC=GetDC(MainWin);

	while (1){
		if (ThreadIsDying){
			ReleaseDC(MainWin, DC);
			return 0;
		}

		if (y >= MaxY){
			if (j==false){
				char buf[180];
				sprintf(buf, "Time/Secs is%f", ((float)GetTickCount()-t)/1000);
				SetWindowText(MainWin, buf);
			}
			j = true;

			Sleep(1000);
		}else{
			for (x = 0; x < MaxX; x++){
				SetPoint(&pixel, x, y, 0);
				SetVectorFromPoints(&direction, &pixel, &ViewPoint);
				RaySet(&Root, rtNormal, &ViewPoint, &direction, 0, 0);
				RayHit(&Root, &shade);

				Line[x] = GetRGBFromColor(&shade);
			}

			EnterCriticalSection(&CriticalSection);

			TempDC=CreateCompatibleDC(DC);
			SelectObject(TempDC, hSceneBmp);

			for (x = 0; x < MaxX; x++){
				SetPixel(TempDC, x, y, Line[x]);
			}

			BitBlt(DC, 0, y, MaxX - 1, 1, TempDC, 0, y, SRCCOPY);

			DeleteDC(TempDC);

			y++;

			LeaveCriticalSection(&CriticalSection);
		}
	}
	return 0;
}
//---------------------------------------------------------------------------
void DestroyRayTracer()
{
	ThreadIsDying++;
	WaitForSingleObject(hThread, 10000);

	CloseHandle(hThread);

	DeleteCriticalSection(&CriticalSection);

	DeleteObject(hSceneBmp); hSceneBmp = NULL;
}
//---------------------------------------------------------------------------
// WndProc
//---------------------------------------------------------------------------
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message){
		case WM_PAINT:
		if (hSceneBmp){
			BITMAP bitMap;
			PAINTSTRUCT ps;
			POINT ptSize,ptOrg;
			HDC DC,MemDC;

			DC = BeginPaint(hwnd,&ps);

			EnterCriticalSection(&CriticalSection);

			GetObject(hSceneBmp, sizeof(BITMAP), &bitMap);

			MemDC=CreateCompatibleDC(DC);
			SelectObject(MemDC, hSceneBmp);
			SetMapMode(MemDC, GetMapMode(DC));

			ptSize.x = bitMap.bmWidth;
			ptSize.y = bitMap.bmHeight;
			DPtoLP(DC, &ptSize,1);

			ptOrg.x = 0;
			ptOrg.y = 0;
			DPtoLP(MemDC, &ptOrg, 1);

			BitBlt(DC, 0, 0, ptSize.x, ptSize.y, MemDC, ptOrg.x, ptOrg.y, SRCCOPY);

			DeleteDC(MemDC);

			LeaveCriticalSection(&CriticalSection);

			EndPaint(hwnd, &ps);
		}
		return 0;

		case WM_DESTROY:
			PostQuitMessage(0);
		return 0;

	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}

