//------------------------------------
//  (c) Reliable Software, 1996, 97
//------------------------------------

// Sphere
void SphereSet(SPHERE *S, RPOINT *center, double radius, OPTICALPROP *pr)
{
  S->Obj.Type = OBJECT_SPHERE;
  S->Obj.OpticalProperty = *pr;
  S->Center   = *center;
  S->Radius   = radius;
  S->R2       = radius*radius;
}

// Vector normal at intersection
VECTOR SphereNormal(SPHERE *S, RPOINT *intersect)
{
  VECTOR result;

  SetVectorFromPoints(&result, intersect, &S->Center);
  NormalizeVector(&result);
  return result;
}

bool SphereIntersect(SPHERE *S, RAY *pRay)
{
  VECTOR oc;
  double oc2, dist2, tca, hc2;

  // ray = _origin + t * _direction;
  // Sphere: (v - _center )^2 = r * r;
  // intersection:
  // ( _origin - _center + t * _direction ) ^2 = r^2;
  // (_o-_c)^2 + 2 * t * ( (_o-_c) . _direction ) + t^2  = r^2;
  // _direction is normalized to 1
  // disc = 4 * ((_o-_c) . _direction ) - 4 * (_o-_c)^2
  // t = (- 2 * ((_o-_c) . _direction) + sqrt ( disc ) ) / 2

  // origin to center vector
  SetVectorFromPoints(&oc, &S->Center, &pRay->Origin);

  // square distance
  oc2 = GetDotProductOfVectors(&oc, &oc);
  dist2 = oc2 - S->R2;

  if ( dist2 <= EPSILON && dist2 >= -EPSILON ) // on the surface
      return false;

  // closest approach along the ray
  tca = GetDotProductOfVectors(&oc, &pRay->Direction);

  if (dist2 > EPSILON && tca < 0.0) // outside and pointing away
      return false;

  // half cord squared
  hc2 = S->R2 - oc2 + tca * tca;

  if (hc2 < 0.0) // no real solution
      return false;

  if (dist2 > EPSILON) // outside
      RayRecord(pRay,tca - sqrt ( hc2 ), (VISIBLEOBJECT*)S);
  else
      RayRecord(pRay,tca + sqrt ( hc2 ), (VISIBLEOBJECT*)S);

  return true;
}

Shader *SphereCreateShader(SPHERE *S, RAY *pRay)
{
  Shader *pShader;

//  pShader = (Shader*)malloc(sizeof(Shader));
  pShader = (Shader*)HeapAlloc(g_hHeap, 0, sizeof(Shader));
  ShaderSet(pShader, SHADER_PHONG, pRay, (VISIBLEOBJECT*)S);
  return pShader;
}


