#if !defined VECTOR_H
#define VECTOR_H

const double EPSILON = 0.001;

// Things you can do with a point:
// Shift it by adding or subtracting a vector
typedef struct{
	double x,y,z;
}RPOINT;

void SetPoint(RPOINT *P,double _x, double _y, double _z)
{
  P->x = _x;
  P->y = _y;
  P->z = _z;
}
//---------------------------------------------------------------------------
// Things you can do with a vector:
// Create it from two points (endpoint and startpoint)
// Add or subtract another vector (inherited from RPOINT)
// Stretch (or shrink) by multiplying it by an number
// Multiply by another vector to obtain their scalar (dot) product
// Calculate its length
// Normalize (stretch so that its length is one)
typedef struct{
	double x, y, z;
}VECTOR;

void SetVector(VECTOR *V,double _x, double _y, double _z)
{
  V->x = _x;
  V->y = _y;
  V->z = _z;
}

void SetVectorFromPoints(VECTOR *V,RPOINT *endpoint,RPOINT *origin)
{
  V->x=endpoint->x-origin->x;
  V->y=endpoint->y-origin->y;
  V->z=endpoint->z-origin->z;
}

void NormalizeVector(VECTOR *V)
{
  double len = sqrt (V->x*V->x + V->y*V->y + V->z*V->z);
  if (len > EPSILON)
  {
    double invLen = 1 / len;
    V->x *= invLen;
    V->y *= invLen;
    V->z *= invLen;
  }
  else
  {
    V->x = 0.0;
    V->y = 0.0;
    V->z = 0.0;
  }
}

double GetDotProductOfVectors(VECTOR *V1,VECTOR *V2)
{
  return V1->x*V2->x + V1->y*V2->y + V1->z*V2->z;
}

void MultiplyVector(VECTOR *V,double By)
{
  V->x *= By;
  V->y *= By;
  V->z *= By;
}

void AddVectorToPoint(RPOINT *V,VECTOR *P)
{
  V->x += P->x;
  V->y += P->y;
  V->z += P->z;
}

void AddVectorToVector(VECTOR *V,VECTOR *P)
{
  V->x += P->x;
  V->y += P->y;
  V->z += P->z;
}

bool IsVectorNonZero(VECTOR *V)
{
  return V->x!=0.0 || V->y!=0.0 || V->z!=0.0;
}
//---------------------------------------------------------------------------
typedef struct{
  double r, g, b;
}COLOR;

void SetColor(COLOR *Col,double red, double green, double blue)
{
  Col->r = red;
  Col->g = green;
  Col->b = blue;
}

COLORREF GetRGBFromColor(COLOR *C)
{
  int r = (int)ceil(C->r * 255.0), g=(int)ceil(C->g * 255.0), b=(int)ceil(C->b * 255.0);
  return RGB(min(r,255),min(g,255),min(b,255));
}

const double normFactor = 1.0/255.0;

void SetColorFromRGB(COLOR *C,COLORREF rgb)
{
  C->r = GetRValue(rgb) * normFactor;
  C->g = GetGValue(rgb) * normFactor;
  C->b = GetBValue(rgb) * normFactor;
}

void MultiplyColorByColor(COLOR *C,COLOR *By)
{
  C->r *= By->r;
  C->g *= By->g;
  C->b *= By->b;
}

void MultiplyColor(COLOR *C,double intens)
{
  C->r *= intens;
  C->g *= intens;
  C->b *= intens;
}

void AddColorToColor(COLOR *C,COLOR *C2)
{
  C->r += C2->r;
  C->g += C2->g;
  C->b += C2->b;
}

#endif
